package se.chalmers.paragon;

import java.util.HashSet;

/**
 * 
 * <p>
 * Class that functions as a factory for Actors. An actor is at runtime
 * represented as an instance of this class. There are two possible instances of
 * the actor: a concrete actor that represents an entity, and an actor variable
 * that can be used in the construction of lock clauses or policies.
 * </p>
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class Actor {
	
	private static HashSet<String> names = new HashSet<String>();

	/**
	 * Actors cannot be created outside this package, except via the public
	 * factory methods.
	 */
	protected Actor() {
	}

	/**
	 * Constructs a new concrete actor
	 * 
	 * @param name
	 *            The name of this concrete actor
	 * 
	 * @return A new instance of the {@link se.chalmers.paragon.ConcreteActor}
	 *         class
	 */
	public static ConcreteActor newConcreteActor(String name) {
		if (names.contains(name))
			throw new UniqueNameError("Actor name " + name + " is not unique.");
		names.add(name);
		return new ConcreteActor(name);
	}
	
	public static ConcreteActor freshActor(String base) {
		int i = 0;
		String name = base;
		while (names.contains(name)) {
			name = base + i;
			i += 1;
		}
		return newConcreteActor(name);
	}

	// TODO We could also have only one instance for each argument 'var', that
	// would give a different comparison (i.e. not overwriting the equals
	// method) but unknown how this could affect the containment check.

	/**
	 * Constructs a new actor variable
	 * 
	 * @param var
	 *            The identifier for this variable in its current clause.
	 * @return A new instance of the {@link se.chalmers.paragon.ActorVariable}
	 *         class
	 */
	public static ActorVariable newActorVariable(int var) {
		return new ActorVariable(var);
	}

}
